/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "pwmd-error.h"
#include <gcrypt.h>
#include "mutex.h"
#include "rcfile.h"
#include "util-misc.h"
#include "common.h"
#include "util-slist.h"
#include "util-string.h"
#include "mem.h"

#define DEFAULT_PINENTRY_TIMEOUT "30"
#define DEFAULT_CACHE_TIMEOUT "600"
#define DEFAULT_KEEPALIVE_INTERVAL "60"

#define INVALID_VALUE(file, line) do {					\
    if (file)								\
      log_write(_("%s(%i): invalid value for parameter."), file, line); \
  } while (0);

enum
{
  PARAM_INT, PARAM_CHARP, PARAM_LONG, PARAM_LONGLONG, PARAM_CHARPP,
  PARAM_BOOL, PARAM_ULONG, PARAM_ULONGLONG
};

static struct config_params_s
{
  char *name;
  int type;
  char *value;
} config_params[] = {
  {  "backup",			PARAM_BOOL,		"true"},
  {  "socket_path",		PARAM_CHARP,		NULL},
  {  "socket_perms",		PARAM_CHARP,		NULL},
  {  "passphrase",		PARAM_CHARP,		NULL},
  {  "passphrase_file",		PARAM_CHARP,		NULL},
  {  "agent_env_file",		PARAM_CHARP,		NULL},
  {  "log_path",		PARAM_CHARP,		"~/.pwmd/log"},
  {  "enable_logging",		PARAM_BOOL,		"0"},
  {  "log_keepopen",            PARAM_BOOL,             "true"},
  {  "log_level",		PARAM_INT,		"0"},
  {  "disable_mlockall",	PARAM_BOOL,		"true"},
  {  "cache_timeout",		PARAM_INT,		DEFAULT_CACHE_TIMEOUT},
  {  "cache_push",		PARAM_CHARPP,		NULL},
  {  "disable_list_and_dump",	PARAM_BOOL,		"false"},
  {  "recursion_depth",		PARAM_INT,		"100"},
  {  "syslog",			PARAM_BOOL,		"false"},
  {  "xfer_progress",		PARAM_INT,		"8196"},
  {  "allowed",			PARAM_CHARPP,		NULL},
  {  "keyparam",		PARAM_CHARP,		"(genkey (rsa (nbits 4:2048)))"},
  {  "cipher",			PARAM_CHARP,		"aes256"},
  {  "kill_scd",		PARAM_BOOL,		"false"},
  {  "cipher_iterations",	PARAM_ULONGLONG,	"0"},
  {  "cipher_progress",		PARAM_LONG,		DEFAULT_ITERATION_PROGRESS},
  {  "priority",		PARAM_INT,		INVALID_PRIORITY},
  {  "keepalive_interval",      PARAM_INT,              DEFAULT_KEEPALIVE_INTERVAL},
  {  "tcp_port",		PARAM_INT,		"6466"},
  {  "enable_tcp",		PARAM_BOOL,		"false"},
  {  "tcp_require_key",		PARAM_BOOL,		"false"},
  {  "tcp_wait",		PARAM_INT,		"0"},
  {  "tcp_bind",		PARAM_CHARP,		"any"},
  {  "tcp_interface",           PARAM_CHARP,            NULL},
  {  "tls_timeout",             PARAM_INT,              "300"},
  {  "tls_cipher_suite",	PARAM_CHARP,		"SECURE256"},
  {  "tls_access",		PARAM_CHARPP,		NULL},
  {  "pinentry_path",           PARAM_CHARP,            PINENTRY_PATH},
  {  "pinentry_timeout",        PARAM_INT,              DEFAULT_PINENTRY_TIMEOUT},
  {  "use_agent",               PARAM_BOOL,             "false"},
  {  "require_save_key",        PARAM_BOOL,             "true"},
  { NULL,			0,			NULL},
};

struct config_param_s
{
  char *name;
  int type;
  union
  {
    int itype;
    char *cptype;
    char **cpptype;
    long ltype;
    long long lltype;
    unsigned long ultype;
    unsigned long long ulltype;
  } value;
};

static struct config_section_s *config_find_section (struct slist_s *config,
					      const char *name);
static int new_param (struct config_section_s *section, const char *filename,
		      int lineno, const char *name, const char *value,
		      int type);
static void free_section (struct config_section_s *s);
static int set_defaults (struct slist_s **config);

static void
section_remove_param (struct config_section_s *section, const char *name)
{
  unsigned i, t = slist_length (section->params);

  for (i = 0; i < t; i++)
    {
      struct config_param_s *p = slist_nth_data (section->params, i);

      if (!p)
	continue;

      if (!strcmp (p->name, name))
	{
	  switch (p->type)
	    {
	    case PARAM_CHARP:
	      xfree (p->value.cptype);
	      break;
	    case PARAM_CHARPP:
	      strv_free (p->value.cpptype);
	      break;
	    }

	  section->params = slist_remove (section->params, p);
	  xfree (p->name);
	  xfree (p);
	  break;
	}
    }
}

void
config_clear_keys ()
{
  MUTEX_LOCK (&rcfile_mutex);
  unsigned i, t = slist_length (global_config);

  for (i = 0; i < t; i++)
    {
      struct config_section_s *s = slist_nth_data (global_config, i);
      if (!s)
	continue;

      section_remove_param (s, "passphrase");
    }

  MUTEX_UNLOCK (&rcfile_mutex);
}

static struct config_param_s *
config_has_param (struct config_section_s *s, const char *what)
{
  unsigned i, t = slist_length (s->params);

  for (i = 0; i < t; i++)
    {
      struct config_param_s *p = slist_nth_data (s->params, i);
      if (!p)
	break;

      if (!strcmp (p->name, what))
	return p;
    }

  return NULL;
}

static struct config_param_s *
config_get_param (struct slist_s *config,
		  const char *section, const char *what, int *exists)
{
  unsigned i, t = slist_length (config);

  *exists = 0;

  for (i = 0; i < t; i++)
    {
      struct config_param_s *p;
      struct config_section_s *s = slist_nth_data (config, i);

      if (!s)
	break;

      if (strcmp (s->name, section))
	continue;

      p = config_has_param (s, what);
      if (!p)
	return NULL;

      *exists = 1;
      return p;
    }

  return NULL;
}

static struct config_section_s *
new_section (struct slist_s **config, const char *name)
{
  struct slist_s *tmp;
  struct config_section_s *s = xcalloc (1, sizeof (struct config_section_s));

  if (!s)
    return NULL;

  s->name = str_dup (name);
  if (!s->name)
    {
      log_write ("%s", pwmd_strerror (ENOMEM));
      xfree (s);
      return NULL;
    }

  tmp = slist_append (*config, s);
  if (!tmp)
    {
      log_write ("%s", pwmd_strerror (ENOMEM));
      xfree (s->name);
      xfree (s);
      return NULL;
    }

  *config = tmp;
  return s;
}

int
config_set_string_param (struct slist_s **config, const char *section,
			 const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_CHARP);
}

char *
config_get_string_param (struct slist_s *config, const char *section,
			 const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists && p->value.cptype ? str_dup (p->value.cptype) : NULL;
}

int
config_set_int_param (struct slist_s **config, const char *section,
		      const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_INT);
}

int
config_get_int_param (struct slist_s *config, const char *section,
		      const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists ? p->value.itype : -1;
}

int
config_set_bool_param (struct slist_s **config, const char *section,
		       const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_BOOL);
}

int
config_get_bool_param (struct slist_s *config, const char *section,
		       const char *what, int *exists)
{
  return config_get_int_param (config, section, what, exists);
}

int
config_set_long_param (struct slist_s **config, const char *section,
		       const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_LONG);
}

unsigned long
config_get_ulong_param (struct slist_s *config, const char *section,
			const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists ? p->value.ultype : 0;
}

long
config_get_long_param (struct slist_s *config, const char *section,
		       const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists ? p->value.ltype : -1;
}

int
config_set_longlong_param (struct slist_s **config, const char *section,
			   const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_LONGLONG);
}

long long
config_get_longlong_param (struct slist_s *config,
			   const char *section, const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists ? p->value.lltype : -1;
}

unsigned long long
config_get_ulonglong_param (struct slist_s *config,
			    const char *section,
			    const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists ? p->value.ulltype : 0;
}

int
config_set_list_param (struct slist_s **config, const char *section,
		       const char *name, const char *value)
{
  struct config_section_s *s = config_find_section (*config, section);

  if (!s)
    {
      s = new_section (config, section);
      if (!s)
	return 1;
    }

  return new_param (s, NULL, 0, name, value, PARAM_CHARPP);
}

char **
config_get_list_param (struct slist_s *config, const char *section,
		       const char *what, int *exists)
{
  struct config_param_s *p = config_get_param (config, section, what, exists);
  return *exists && p->value.cpptype ? strv_dup (p->value.cpptype) : NULL;
}

char *
config_get_string (const char *section, const char *what)
{
  char *val = NULL;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_string_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_string_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

char **
config_get_list (const char *section, const char *what)
{
  char **val = NULL;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_list_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_list_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

int
config_get_integer (const char *section, const char *what)
{
  int val = 0;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_int_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_int_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

long long
config_get_longlong (const char *section, const char *what)
{
  long long val = 0;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_longlong_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_longlong_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

unsigned long long
config_get_ulonglong (const char *section, const char *what)
{
  unsigned long long val = 0;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_ulonglong_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_ulonglong_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

long
config_get_long (const char *section, const char *what)
{
  long val = 0;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_long_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_long_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

unsigned long
config_get_ulong (const char *section, const char *what)
{
  unsigned long val = 0;
  const char *where = section ? section : "global";
  int exists = 0;

  MUTEX_LOCK (&rcfile_mutex);
  val = config_get_ulong_param (global_config, where, what, &exists);
  if (!exists && strcmp (section ? section : "", "global"))
    val = config_get_ulong_param (global_config, "global", what, &exists);

  MUTEX_UNLOCK (&rcfile_mutex);
  return val;
}

int
config_get_boolean (const char *section, const char *what)
{
  return config_get_integer (section, what);
}

char *
config_get_value (const char *section, const char *what)
{
  const char *where = section ? section : "global";
  int exists = 0;
  int i;
  int ival;
  long lval;
  long long llval;
  unsigned long ulval;
  unsigned long long ullval;
  char *cpval;
  char **cppval;
  char *result = NULL;

  MUTEX_LOCK (&rcfile_mutex);

  for (i = 0; config_params[i].name; i++)
    {
      if (!strcmp (config_params[i].name, what))
	{
	  switch (config_params[i].type)
	    {
	    case PARAM_BOOL:
	    case PARAM_INT:
	      ival = config_get_int_param (global_config, where, what,
					   &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		ival = config_get_int_param (global_config, "global", what,
					     &exists);
	      result = str_asprintf ("%i", ival);
	      break;
	    case PARAM_CHARP:
	      cpval = config_get_string_param (global_config, where, what,
					       &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		cpval =
		  config_get_string_param (global_config, "global", what,
					   &exists);
	      result = cpval;
	      break;
	    case PARAM_LONG:
	      lval = config_get_long_param (global_config, where, what,
					    &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		lval = config_get_long_param (global_config, "global", what,
					      &exists);
	      result = str_asprintf ("%li", lval);
	      break;
	    case PARAM_ULONG:
	      ulval = config_get_ulong_param (global_config, where, what,
					      &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		ulval = config_get_ulong_param (global_config, "global", what,
						&exists);
	      result = str_asprintf ("%lu", ulval);
	      break;
	    case PARAM_LONGLONG:
	      llval = config_get_longlong_param (global_config, where, what,
						 &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		llval = config_get_longlong_param (global_config, "global",
						   what, &exists);
	      result = str_asprintf ("%lli", llval);
	      break;
	    case PARAM_ULONGLONG:
	      ullval = config_get_ulonglong_param (global_config, where, what,
						   &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		ullval = config_get_ulonglong_param (global_config, "global",
						     what, &exists);
	      result = str_asprintf ("%llu", ullval);
	      break;
	    case PARAM_CHARPP:
	      cppval = config_get_list_param (global_config, where, what,
					      &exists);
	      if (!exists && strcmp (section ? section : "", "global"))
		cppval = config_get_list_param (global_config, "global", what,
						&exists);

	      if (cppval)
		result = strv_join (",", cppval);

	      strv_free (cppval);
	      break;
	    }
	}
    }

  MUTEX_UNLOCK (&rcfile_mutex);
  return result;
}

static int
set_defaults (struct slist_s **config)
{
  char *s;
  char **list;
  int exists;
  int i;

  for (i = 0; config_params[i].name; i++)
    {
      switch (config_params[i].type)
	{
	case PARAM_BOOL:
	  config_get_bool_param (*config, "global", config_params[i].name,
				 &exists);
	  if (!exists)
	    {
	      if (config_set_bool_param
		  (config, "global", config_params[i].name,
		   config_params[i].value))
		goto fail;
	    }
	  break;
	case PARAM_INT:
	  config_get_int_param (*config, "global", config_params[i].name,
				&exists);
	  if (!exists)
	    {
	      if (config_set_int_param
		  (config, "global", config_params[i].name,
		   config_params[i].value))
		goto fail;
	    }
	  break;
	case PARAM_CHARP:
	  s = config_get_string_param (*config, "global",
					config_params[i].name, &exists);
	  xfree (s);
	  if (!exists && config_params[i].value)
	    {
	      if (config_set_string_param (config, "global",
					   config_params[i].name,
					   config_params[i].value))
		goto fail;
	    }
	  break;
	case PARAM_CHARPP:
	  list = config_get_list_param (*config, "global",
					config_params[i].name, &exists);
	  strv_free (list);
	  if (!exists && config_params[i].value)
	    {
	      if (config_set_list_param (config, "global",
					 config_params[i].name,
					 config_params[i].value))
		goto fail;
	    }
	  break;
	case PARAM_LONG:
	  config_get_long_param (*config, "global", config_params[i].name,
				 &exists);
	  if (!exists)
	    {
	      if (config_set_long_param
		  (config, "global", config_params[i].name,
		   config_params[i].value))
		goto fail;
	    }
	  break;
	case PARAM_LONGLONG:
	  config_get_longlong_param (*config, "global", config_params[i].name,
				     &exists);
	  if (!exists)
	    {
	      if (config_set_longlong_param (config, "global",
					     config_params[i].name,
					     config_params[i].value))
		goto fail;
	    }
	  break;
	}
    }

  list = config_get_list_param (*config, "global", "allowed", &exists);
  strv_free (list);
  if (!exists)
    {
      if (config_set_list_param
	  (config, "global", "allowed", get_username ()))
	goto fail;
    }

  max_recursion_depth = config_get_int_param (*config, "global",
					      "recursion_depth", &exists);
  disable_list_and_dump = config_get_bool_param (*config, "global",
						 "disable_list_and_dump",
						 &exists);
#ifdef HAVE_MLOCKALL
  disable_mlock =
    config_get_bool_param (*config, "global", "disable_mlockall", &exists);
#endif
  return 0;

fail:
  return 1;
}

static struct config_section_s *
config_find_section (struct slist_s *config, const char *name)
{
  unsigned i, t = slist_length (config);

  for (i = 0; i < t; i++)
    {
      struct config_section_s *s = slist_nth_data (config, i);

      if (!strcmp (s->name, name))
	return s;
    }

  return NULL;
}

/* Append a new parameter to the list of parameters for a file
 * section. When an existing parameter of the same name exists, its
 * value is updated.
 */
static int
new_param (struct config_section_s *section, const char *filename, int lineno,
	   const char *name, const char *value, int type)
{
  struct config_param_s *param = NULL;
  struct slist_s *tmp;
  char *e;
  unsigned i, t = slist_length (section->params);
  int dup = 0;

  for (i = 0; i < t; i++)
    {
      struct config_param_s *p = slist_nth_data (section->params, i);
      if (!p)
	break;

      if (!strcmp (name, p->name))
	{
	  param = p;
	  dup = 1;
	  break;
	}
    }

  if (!param)
    {
      param = xcalloc (1, sizeof (struct config_param_s));
      if (!param)
	{
	  log_write ("%s", pwmd_strerror (ENOMEM));
	  return 1;
	}

      param->name = str_dup (name);
      if (!param->name)
	{
	  xfree (param);
	  log_write ("%s", pwmd_strerror (ENOMEM));
	  return 1;
	}
    }

  param->type = type;

  switch (type)
    {
    case PARAM_BOOL:
      if (!strcasecmp (value, "no") || !strcasecmp (value, "0")
	  || !strcasecmp (value, "false"))
	param->value.itype = 0;
      else if (!strcasecmp (value, "yes") || !strcasecmp (value, "1")
	       || !strcasecmp (value, "true"))
	param->value.itype = 1;
      else
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      param->type = PARAM_INT;
      break;
    case PARAM_CHARP:
      xfree (param->value.cptype);
      param->value.cptype = value ? str_dup (value) : NULL;
      if (value && !param->value.cptype)
	{
	  log_write ("%s", pwmd_strerror (ENOMEM));
	  goto fail;
	}
      break;
    case PARAM_CHARPP:
      strv_free (param->value.cpptype);
      param->value.cpptype = value ? str_split (value, ",", 0) : NULL;
      if (value && !param->value.cpptype)
	{
	  log_write ("%s", pwmd_strerror (ENOMEM));
	  goto fail;
	}
      break;
    case PARAM_INT:
      param->value.itype = strtol (value, &e, 10);
      if (e && *e)
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      break;
    case PARAM_LONG:
      param->value.ltype = strtol (value, &e, 10);
      if (e && *e)
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      break;
    case PARAM_LONGLONG:
      param->value.lltype = strtoll (value, &e, 10);
      if (e && *e)
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      break;
    case PARAM_ULONGLONG:
      param->value.ulltype = strtoull (value, &e, 10);
      if (e && *e)
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      break;
    case PARAM_ULONG:
      param->value.ultype = strtoul (value, &e, 10);
      if (e && *e)
	{
	  INVALID_VALUE (filename, lineno);
	  goto fail;
	}
      break;
    }

  if (dup)
    return 0;

  tmp = slist_append (section->params, param);
  if (!tmp)
    {
      log_write ("%s", pwmd_strerror (ENOMEM));
      goto fail;
    }

  section->params = tmp;
  return 0;

fail:
  xfree (param->name);
  xfree (param);
  return 1;
}

struct slist_s *
config_parse (const char *filename)
{
  struct slist_s *tmpconfig = NULL, *tmp;
  struct config_section_s *cur_section = NULL;
  char buf[LINE_MAX];
  char *s;
  int lineno = 1;
  int have_global = 0;
  FILE *fp = fopen (filename, "r");

  if (!fp)
    {
      log_write ("%s: %s", filename,
		 pwmd_strerror (gpg_error_from_errno (errno)));

      if (errno != ENOENT)
	return NULL;

      log_write (_("Using defaults!"));
      goto defaults;
    }

  for (; (s = fgets (buf, sizeof (buf), fp)); lineno++)
    {
      char line[LINE_MAX] = { 0 };
      size_t len = 0;

      if (*s == '#')
	continue;

      s = str_chomp (s);
      for (; s && *s; s++)
	{
	  int match = 0;

	  /* New file section. */
	  if (*s == '[')
	    {
	      struct config_section_s *section;
	      char *p = strchr (++s, ']');

	      if (!p)
		{
		  log_write (_("%s(%i): unbalanced braces"), filename,
			     lineno);
		  goto fail;
		}

	      len = strlen (s) - strlen (p);
	      memcpy (line, s, len);
	      line[len] = 0;

	      section = config_find_section (tmpconfig, line);
	      if (section)
		{
		  log_write (_("%s(%i): section '%s' already exists!"),
			     filename, lineno, line);
		  goto fail;
		}

	      if (!strcmp (line, "global"))
		have_global = 1;

	      section = xcalloc (1, sizeof (struct config_section_s));
	      section->name = str_dup (line);

	      if (cur_section)
		{
		  tmp = slist_append (tmpconfig, cur_section);
		  if (!tmp)
		    {
		      log_write ("%s", pwmd_strerror (ENOMEM));
		      goto fail;
		    }

		  tmpconfig = tmp;
		}

	      cur_section = section;
	      break;
	    }

	  if (!cur_section)
	    {
	      log_write (_("%s(%i): parameter outside of section!"), filename,
			 lineno);
	      goto fail;
	    }

	  /* Parameters for each section. */
	  for (int m = 0; config_params[m].name; m++)
	    {
	      size_t len = strlen (config_params[m].name);

	      if (!strncmp (s, config_params[m].name, len))
		{
		  char *p = s + len;

		  while (*p && *p == ' ')
		    p++;

		  if (!*p || *p != '=')
		    continue;

		  p++;
		  while (*p && isspace (*p))
		    p++;

		  s[len] = 0;
		  if (new_param (cur_section, filename, lineno, s, p,
				 config_params[m].type))
		    goto fail;

		  match = 1;
		  break;
		}
	    }

	  if (!match)
	    {
	      log_write (_("%s(%i): unknown parameter"), filename, lineno);
	      goto fail;
	    }

	  break;
	}
    }

  if (cur_section)
    {
      tmp = slist_append (tmpconfig, cur_section);
      if (!tmp)
	{
	  log_write ("%s", pwmd_strerror (ENOMEM));
	  goto fail;
	}

      cur_section = NULL;
      tmpconfig = tmp;
    }

  if (!have_global)
    log_write (_
	       ("WARNING: %s: could not find a [global] configuration section!"),
	       filename);

defaults:
  if (set_defaults (&tmpconfig))
    goto fail;

  if (fp)
    fclose(fp);

  return tmpconfig;

fail:
  if (fp)
    fclose (fp);

  config_free (tmpconfig);
  free_section (cur_section);
  return NULL;
}

static void
free_section (struct config_section_s *s)
{
  if (!s)
    return;

  for (;;)
    {
      struct config_param_s *p = slist_nth_data (s->params, 0);

      if (!p)
	break;

      section_remove_param (s, p->name);
    }

  s->params = NULL;
  xfree (s->name);
  s->name = NULL;
}

void
config_free (struct slist_s *config)
{
  for (;;)
    {
      struct config_section_s *s = slist_nth_data (config, 0);
      if (!s)
	break;

      free_section (s);
      config = slist_remove (config, s);
      xfree (s);
    }
}
